
# This file was generated by 'versioneer.py' (0.21) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2022-01-27T10:07:19+0000",
 "dirty": false,
 "error": null,
 "full-revisionid": "9a8df8ff19a0a71eb7f0973a91fe57458c3c1eb1",
 "version": "v0.1.21"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
